from skyfield.api import load
from skyfield.magnitudelib import planetary_magnitude

def test_magnitudes():
    ts = load.timescale()
    #t = ts.utc(1995, 5, 22)  # Rings edge-on from Earth.
    t = ts.utc(2021, 10, 4)
    eph = load('de421.bsp')
    earth = eph['earth']
    names = [
        'mercury', 'venus', 'mars', 'jupiter barycenter',
        'saturn barycenter', 'uranus barycenter', 'neptune barycenter',
    ]
    e = earth.at(t)
    positions = [e.observe(eph[name]) for name in names]
    magnitudes = [planetary_magnitude(position) for position in positions]
    assert ['%.3f' % m for m in magnitudes] == [
        '2.393', '-4.278', '1.592', '-2.693', '0.508', '5.701', '7.690',
    ]

    t = ts.tt_jd([t.tt, t.tt])
    e = earth.at(t)
    positions = [e.observe(eph[name]) for name in names]
    magnitudes = [planetary_magnitude(position) for position in positions]
    assert [['%.3f' % m for m in mm] for mm in magnitudes] == [
        ['2.393', '2.393'],
        ['-4.278', '-4.278'],
        ['1.592', '1.592'],
        ['-2.693', '-2.693'],
        ['0.508', '0.508'],
        ['5.701', '5.701'],
        ['7.690', '7.690'],
    ]



"""Low-level tests of the almanac search routines."""

import numpy as np

from assay import assert_raises
from skyfield.api import load
from skyfield.searchlib import find_discrete, find_maxima

bump = 1e-5
epsilon = 1e-10

def make_t():
    ts = load.timescale()
    t0 = ts.tt_jd(0)
    t1 = ts.tt_jd(1)
    return t0, t1

def make_stairstep_f(steps):
    """Return a function that increases by one at each of several `steps`."""
    def f(t):
        # For each time, sum how many of the values in `steps` it surpasses.
        return np.greater_equal.outer(t.tt, steps).sum(axis=1)
    f.step_days = 0.3
    return f

def is_close(value, expected):
    return (abs(value - expected) < epsilon).all()

def test_exception_if_step_days_is_missing():
    def f(t):
        return t.J > 0.0
    t0, t1 = make_t()
    with assert_raises(AttributeError, 'missing a "step_days" attribute'):
        find_discrete(t0, t1, f, epsilon)

def test_find_discrete_that_finds_nothing():
    t0, t1 = make_t()
    f = make_stairstep_f([-0.1, +1.1])
    t, y = find_discrete(t0, t1, f, epsilon)
    assert not len(t.tt)
    assert not len(y)

def test_find_discrete_near_left_edge():
    t0, t1 = make_t()
    f = make_stairstep_f([bump, 0.5])
    t, y = find_discrete(t0, t1, f, epsilon)
    assert is_close(t.tt, (bump, 0.5))
    assert list(y) == [1, 2]

def test_find_discrete_near_right_edge():
    t0, t1 = make_t()
    f = make_stairstep_f([0.5, 1.0 - bump])
    t, y = find_discrete(t0, t1, f, epsilon)
    assert is_close(t.tt, (0.5, 1.0 - bump))
    assert list(y) == [1, 2]

def test_find_discrete_with_a_barely_detectable_jag_right_at_zero():
    t0, t1 = make_t()
    f = make_stairstep_f([0.5, 0.5 + 3.1 * epsilon])
    t, y = find_discrete(t0, t1, f, epsilon)
    assert is_close(t.tt, (0.5, 0.5 + 3.1 * epsilon))
    assert list(y) == [1, 2]

def DISABLED_test_find_discrete_with_a_sub_epsilon_jag_right_at_zero():
    t0, t1 = make_t()
    f = make_stairstep_f([0.5, 0.5 + 0.99 * epsilon])

    # We hard-code num=12, just in case the default ever changes to
    # another value that might not trigger the symptom.
    t, y = find_discrete(t0, t1, f, epsilon, 12)

    # Note that we always return the last of several close solutions, so
    # that `y` correctly reflects the new state that persists after the
    # flurry of changes is complete.
    assert is_close(t.tt, (0.5 + 0.99 * epsilon,))
    assert list(y) == [2]

def test_old_rough_period_attribute():
    t0, t1 = make_t()
    f = make_stairstep_f([bump, 0.5])
    del f.step_days
    f.rough_period = 1.0
    t, y = find_discrete(t0, t1, f, epsilon)
    assert is_close(t.tt, (bump, 0.5))
    assert list(y) == [1, 2]

def make_mountain_range_f(peaks):
    """Return a function with local maxima at each of a series of `peaks`."""
    def f(t):
        # For each time, sum how many of the values in `steps` it surpasses.
        return -abs(np.subtract.outer(t.tt, peaks)).min(axis=1)
    f.step_days = 0.3
    return f

def test_finding_enough_maxima():
    # If the step size is small enough, no maxima should be skipped.
    t0, t1 = make_t()
    f = make_mountain_range_f(np.linspace(0.01, 0.99, 30))
    f.step_days = 0.03 / 2.0  # Half of the expected period
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert len(t) == len(y) == 30

def test_finding_maxima_near_edges():
    t0, t1 = make_t()
    f = make_mountain_range_f([bump, 1.0 - bump])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert is_close(t.tt, (bump, 1.0 - bump))
    assert is_close(y, 0.0)

def test_finding_no_maxima_at_all_but_having_near_misses():
    t0, t1 = make_t()
    f = make_mountain_range_f([-bump, 1.0 + bump])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert list(t.tt) == []
    assert list(y) == []

def test_finding_no_maxima_at_all_with_no_near_misses():
    t0, t1 = make_t()
    f = make_mountain_range_f([-100, 101])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert list(t.tt) == []
    assert list(y) == []

def test_that_we_ignore_maxima_slightly_beyond_range():
    t0, t1 = make_t()
    f = make_mountain_range_f([-bump, 1.0 + bump])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert len(t.tt) == 0
    assert len(y) == 0

def test_we_only_get_one_result_for_a_jagged_maximum():
    t0, t1 = make_t()
    almost = 0.49 * epsilon
    f = make_mountain_range_f([0.5 - almost, 0.5 + almost])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    assert len(t.tt) == len(y) == 1

def test_we_get_two_results_for_barely_separate_maxima():
    t0, t1 = make_t()
    enough = 1.51 * epsilon
    f = make_mountain_range_f([0.5 - enough, 0.5 + enough])
    t, y = find_maxima(t0, t1, f, epsilon, 12)
    print(list(t.tt))
    assert len(t.tt) == len(y) == 2



from numpy import cos
from skyfield import framelib
from skyfield.api import SSB, Star, Topos, load, wgs84
from skyfield.constants import AU_M, ERAD
from skyfield.positionlib import Geocentric

def test_radec_and_altaz_angles_and_rates():
    # HORIZONS test data in Skyfield repository: authorities/radec-altaz-rates
    ts = load.timescale()
    t = ts.utc(2021, 2, 3)
    top = wgs84.latlon(35.1844866, 248.347300, elevation_m=2106.9128)
    planets = load('de421.bsp')
    a = (planets['earth'] + top).at(t).observe(planets['mars']).apparent()

    # First, verify RA and declination.

    frame = framelib.true_equator_and_equinox_of_date
    dec, ra, distance, dec_rate, ra_rate, range_rate = (
        a.frame_latlon_and_rates(frame))

    arcseconds = 3600.0
    assert abs((ra.degrees - 40.75836) * arcseconds) < 0.04
    assert abs((dec.degrees - 17.16791) * arcseconds) < 0.005
    assert abs(distance.m - 1.21164331503552 * AU_M) < 120.0

    # Verify RA and declination rates of change.

    assert round(dec_rate.arcseconds.per_hour, 5) == 25.61352
    assert round(ra_rate.arcseconds.per_hour * cos(dec.radians),
                 4) == round(75.15571, 4)  # TODO: get last digit to agree?
    assert abs(range_rate.km_per_s - 16.7926932) < 2e-5

    # Verify altitude and azimuth.

    frame = top
    alt, az, distance, alt_rate, az_rate, range_rate = (
        a.frame_latlon_and_rates(frame))

    assert round(alt.degrees, 4) == 65.2758
    assert round(az.degrees, 4) == 131.8839
    assert abs(distance.m - 1.21164331503552 * AU_M) < 120.0

    # Verify altitude and azimuth rates of change.

    assert abs(range_rate.km_per_s - 16.7926932) < 2e-5
    assert round(alt_rate.arcseconds.per_minute, 2) == 548.66
    assert round(az_rate.arcseconds.per_minute * cos(alt.radians), 2) == 663.55

def test_frame_round_trip():
    # Does a frame's rotation and twist get applied in the right
    # directions?  Let's test whether the position and velocity of an
    # ITRS vector (ERAD,0,0) are restored to the proper orientation.
    top = Topos(latitude_degrees=0, longitude_degrees=0)
    ts = load.timescale()
    t = ts.utc(2020, 11, 27, 15, 34)  # Arbitrary time; LST ~= 20.03.
    p = top.at(t)

    r = p.frame_xyz(framelib.itrs)
    assert max(abs(r.m - [ERAD, 0, 0])) < 4e-8 # meters

    r, v = p.frame_xyz_and_velocity(framelib.itrs)
    assert max(abs(r.m - [ERAD, 0, 0])) < 4e-8 # meters
    assert max(abs(v.km_per_s)) < 3e-15 # km/s

def test_from_frame_method():
    ts = load.timescale()
    t = ts.utc(2020, 11, 27, 15, 34)
    g1 = Geocentric([1,2,3], [4,5,6], t=t)
    r, v = g1.frame_xyz_and_velocity(framelib.itrs) # which we trust: see above

    g2 = Geocentric.from_time_and_frame_vectors(t, framelib.itrs, r, v)
    assert max(abs(g2.xyz.au - [1,2,3])) < 2e-14
    assert max(abs(g2.velocity.au_per_d - [4,5,6])) < 3e-14

    # Make sure original vectors were not harmed (for example, by "+=").
    assert list(g1.xyz.au) == [1,2,3]
    assert list(g1.velocity.au_per_d) == [4,5,6]

def test_frame_without_spin():
    ts = load.timescale()
    t = ts.utc(2020, 11, 27, 15, 34)
    g = Geocentric([1,2,3], [4,5,6], t=t)

    # Simply test whether "None" spin raises an exception in either direction.
    f = framelib.true_equator_and_equinox_of_date
    r, v = g.frame_xyz_and_velocity(f)
    Geocentric.from_time_and_frame_vectors(t, f, r, v)

def test_true_equator_and_equinox_of_date():
    ts = load.timescale()
    t = ts.utc(2025, 7, 2.375)
    alpha_andromeda = Star(
        ra_hours=0.13976888866666667, dec_degrees=29.09082805,
        ra_mas_per_year=135.68, dec_mas_per_year=-162.95, parallax_mas=33.6,
        epoch=ts.J(1991.25),
    )
    dec, ra, _ = SSB.at(t).observe(alpha_andromeda).frame_latlon(
        framelib.mean_equator_and_equinox_of_date
    )
    ra.preference = 'hours'

    # Position from page H2 of the 2025 Astronomical Almanac:
    assert ra.hstr(places=1) == '00h 09m 42.7s'
    assert dec.dstr() == '+29deg 13\' 52.0"'

def test_tirs_at_least_runs():
    # TODO: find an external source for a TIRS vector to test against.
    # For now, just make sure it doesn't raise an exception.
    ts = load.timescale()
    t = ts.utc(2020, 11, 27, 15, 34)
    g = Geocentric([1,2,3], [4,5,6], t=t)
    g.frame_xyz_and_velocity(framelib.tirs)



from skyfield.data import text_pck

TEXT = b"""JPL/PCK
This is ignored.
\\begindata
A = 3
B = (4)
C = ( 5 'Six' )
D_+=( 7,8 )
E = (9 1.0000D1)
E += (11)
E += 12
\\begintext
This is also ignored.
E += (13)
"""

def test_loading():
    lines = TEXT.splitlines(True)
    variables = {}
    text_pck.load(lines, variables)
    assert variables == {
        'A': 3,
        'B': 4,
        'C': [5, 'Six'],
        'D_': [7, 8],
        'E': [9, 10.0, 11, 12],
    }

def test_parsing():
    lines = TEXT.splitlines(True)
    assignments = list(text_pck.parse(lines))
    assert assignments == [
        ('A', b'=', [3]),
        ('B', b'=', [4]),
        ('C', b'=', [5, 'Six']),
        ('D_', b'+=', [7, 8]),
        ('E', b'=', [9, 10.0]),
        ('E', b'+=', [11]),
        ('E', b'+=', [12]),
    ]



import os
from skyfield.api import load, load_file

from assay import assert_raises

def _data_path(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)

# Test file generated with:
# python -m jplephem excerpt 1969/07/29 1969/07/30 de441.bsp de441-1969.bsp

def test_multiple_non_overlapping_segments_per_target():
    ts = load.timescale()
    t = ts.utc(1969, 7, [28, 29, 30, 31])
    eph = load_file(_data_path('./de441-1969.bsp'))

    pluto = eph['pluto barycenter']
    assert type(pluto).__name__ == 'Stack'
    assert len(pluto.segments) == 2

    assert str(pluto.at(t).xyz) == (
        '[[-30.47352052 -30.47318972 -30.47285863 -30.47252727]\n'
        ' [ -0.96671442  -0.96985773  -0.97300104  -0.97614434]\n'
        ' [  8.87919555   8.87811509   8.87703455   8.87595392]] au'
    )

    # Does the Stack correctly offer its .ephemeris to apparent()?

    pluto.at(t).observe(pluto).apparent()

    # TODO: SSB.at(t).observe() fails the above test.

# Verify that ephemeris objects let their segments be edited.

def test_removing_segments_from_ephemeris():
    eph = load('de421.bsp')
    eph.segments = [s for s in eph.segments if s.target in (3, 4)]

    assert len(eph.segments) == 2
    assert 2 not in eph
    assert 3 in eph
    assert eph.codes == {0, 3, 4}
    assert eph.names() == {
        0: ['SOLAR_SYSTEM_BARYCENTER', 'SSB', 'SOLAR SYSTEM BARYCENTER'],
        3: ['EARTH_BARYCENTER', 'EMB', 'EARTH MOON BARYCENTER',
            'EARTH-MOON BARYCENTER', 'EARTH BARYCENTER'],
        4: ['MARS_BARYCENTER', 'MARS BARYCENTER'],
    }

    assert repr(eph) == "<SpiceKernel 'de421.bsp'>"
    assert str(eph) == """\
Segments from kernel file 'de421.bsp':
  JD 2414864.50 - JD 2471184.50  (1899-07-28 through 2053-10-08)
      0 -> 3    SOLAR SYSTEM BARYCENTER -> EARTH BARYCENTER
      0 -> 4    SOLAR SYSTEM BARYCENTER -> MARS BARYCENTER"""

    assert type(eph[4]).__name__ == 'ChebyshevPosition'
    with assert_raises(KeyError, 'is missing 5'):
        eph[5]

def test_adding_segments_to_ephemeris():
    eph = load('de405.bsp')
    eph.segments = [s for s in eph.segments if s.target == 3]

    eph2 = load('de421.bsp')
    eph.segments.extend((s for s in eph2.segments if s.target in (301, 399)))

    assert len(eph.segments) == 3
    assert 2 not in eph
    assert 3 in eph
    assert 301 in eph
    assert eph.codes == {0, 3, 301, 399}
    assert eph.names() == {
        0: ['SOLAR_SYSTEM_BARYCENTER', 'SSB', 'SOLAR SYSTEM BARYCENTER'],
        3: ['EARTH_BARYCENTER', 'EMB', 'EARTH MOON BARYCENTER',
            'EARTH-MOON BARYCENTER', 'EARTH BARYCENTER'],
        301: ['MOON'], 399: ['EARTH'],
    }

    assert repr(eph) == "<SpiceKernel 'de405.bsp' 'de421.bsp'>"
    assert str(eph) == """\
Segments from kernel file 'de405.bsp':
  JD 2305424.50 - JD 2525008.50  (1599-12-08 through 2201-02-19)
      0 -> 3    SOLAR SYSTEM BARYCENTER -> EARTH BARYCENTER
And from kernel file 'de421.bsp':
  JD 2414864.50 - JD 2471184.50  (1899-07-28 through 2053-10-08)
      3 -> 301  EARTH BARYCENTER -> MOON
      3 -> 399  EARTH BARYCENTER -> EARTH"""

    vs = eph[399]
    assert type(vs).__name__ == 'VectorSum'
    assert str(vs) == """\
Sum of 2 vectors:
 'de405.bsp' segment 0 SOLAR SYSTEM BARYCENTER -> 3 EARTH BARYCENTER
 'de421.bsp' segment 3 EARTH BARYCENTER -> 399 EARTH"""

    with assert_raises(KeyError):
        eph[4]

def test_ephemeris_lacking_segments_to_connect_to_barycenter():
    eph = load('de421.bsp')
    eph.segments = [s for s in eph.segments if s.target == 399]
    with assert_raises(KeyError, "Barycenter to the target 'Earth'"):
        eph['Earth']



"""Vector functions and their composition."""

from jplephem.names import target_names as _jpl_code_name_dict
from numpy import max
from .constants import C_AUDAY
from .descriptorlib import reify
from .functions import length_of
from .positionlib import build_position
from .timelib import Time

class VectorFunction(object):
    """Given a time, computes a corresponding position."""

    ephemeris = None

    @reify
    def vector_name(self):
        return type(self).__name__

    @reify
    def center_name(self):
        return _jpl_name(self.center)

    @reify
    def target_name(self):
        return _jpl_name(self.target)

    def __repr__(self):
        return '<{0} {1}>'.format(type(self).__name__, str(self))

    def __str__(self):
        if self.target is self:
            return self.target_name
        return self.arrow_str()

    def arrow_str(self):
        return '{0} {1} -> {2}'.format(
            self.vector_name, self.center_name, self.target_name,
        )

    def __add__(self, other):
        if self.target != other.center:
            if other.target == self.center:
                self, other = other, self
            else:
                raise ValueError(
                    "you can only add two vectors"
                    " if the target where one of the vectors ends"
                    " is the center where the other vector starts"
                )

        self_vfs = getattr(self, 'vector_functions', None) or (self,)
        other_vfs = getattr(other, 'vector_functions', None) or (other,)

        return VectorSum(self.center, other.target, self_vfs + other_vfs)

    def __neg__(self):
        return ReversedVector(self)

    def __sub__(self, other):
        if self.center != other.center:
            raise ValueError(
                "you can only subtract two vectors"
                " if they both start at the same center"
            )

        self_vfs = getattr(self, 'vector_functions', None) or (self,)
        other_vfs = getattr(other, 'vector_functions', None) or (other,)
        other_vfs = tuple(reversed([-vf for vf in other_vfs]))

        return VectorSum(other.target, self.target, other_vfs + self_vfs)

    def at(self, t):
        """At time ``t``, compute the target's position relative to the center.

        If ``t`` is an array of times, then the returned position object
        will specify as many positions as there were times.  The kind of
        position returned depends on the value of the ``center``
        attribute:

        * Solar System Barycenter: :class:`~skyfield.positionlib.Barycentric`
        * Center of the Earth: :class:`~skyfield.positionlib.Geocentric`
        * Anything else: :class:`~skyfield.positionlib.ICRF`

        """
        if not isinstance(t, Time):
            raise ValueError('please provide the at() method with a Time'
                             ' instance as its argument, instead of the'
                             ' value {0!r}'.format(t))
        p, v, gcrs_position, message = self._at(t)
        center = self.center
        position = build_position(p, v, t, center, self.target)
        position._ephemeris = self.ephemeris
        position._observer_gcrs_au = gcrs_position
        position.message = message
        return position

    def _observe_from_bcrs(self, observer):
        if self.center != 0:
            raise ValueError('you can only observe() a body whose vector'
                             ' center is the Solar System Barycenter,'
                             ' but this vector has the center {0}'
                             .format(self.center_name))
        return _correct_for_light_travel_time(observer, self)

class ReversedVector(VectorFunction):
    def __init__(self, vector_function):
        self.center = vector_function.target
        self.target = vector_function.center
        self.vector_function = vector_function

    @reify
    def vector_name(self):
        return 'Reversed ' + self.vector_function.vector_name

    @reify
    def center_name(self):
        return self.vector_function.target_name

    @reify
    def target_name(self):
        return self.vector_function.center_name

    def __neg__(self):
        return self.vector_function

    def _at(self, t):
        p, v, _, message = self.vector_function._at(t)
        return -p, -v, None, message

class VectorSum(VectorFunction):
    def __init__(self, center, target, vector_functions):
        self.center = center
        self.target = target
        self.vector_functions = vector_functions

        # For now, just grab the first ephemeris we can find.
        ephemerides = (segment.ephemeris for segment in vector_functions
                       if segment.ephemeris)
        self.ephemeris = next(ephemerides, None)

    def __str__(self):
        vector_functions = self.vector_functions
        lines = [' ' + segment.arrow_str() for segment in vector_functions]
        return 'Sum of {0} vectors:\n{1}'.format(
            len(vector_functions),
            '\n'.join(lines),
        )

    def __repr__(self):
        return '<Vector{0}>'.format(self)

    def _at(self, t):
        p, v = 0.0, 0.0
        gcrs_position = None
        vfs = self.vector_functions
        for vf in vfs:
            p2, v2, _, message = vf._at(t)
            if vf.center == 399:
                gcrs_position = -p
            p += p2
            v += v2
        if vfs[0].center == 0 and vf.center == 399:
            gcrs_position = p2
        return p, v, gcrs_position, message

def _correct_for_light_travel_time(observer, target):
    """Return a light-time corrected astrometric position and velocity.

    Given an `observer` that is a `Barycentric` position somewhere in
    the solar system, compute where in the sky they will see the body
    `target`, by computing the light-time between them and figuring out
    where `target` was back when the light was leaving it that is now
    reaching the eyes or instruments of the `observer`.

    """
    t = observer.t
    ts = t.ts
    whole = t.whole
    tdb_fraction = t.tdb_fraction

    cposition = observer.xyz.au
    cvelocity = observer.velocity.au_per_d

    tposition, tvelocity, gcrs_position, message = target._at(t)

    distance = length_of(tposition - cposition)
    light_time0 = 0.0
    for i in range(10):
        light_time = distance / C_AUDAY
        delta = light_time - light_time0
        if max(abs(delta), initial=0.0) < 1e-12:
            break

        # We assume a light travel time of at most a couple of days.  A
        # longer light travel time would best be split into a whole and
        # fraction, for adding to the whole and fraction of TDB.
        t2 = ts.tdb_jd(whole, tdb_fraction - light_time)

        tposition, tvelocity, gcrs_position, message = target._at(t2)
        distance = length_of(tposition - cposition)
        light_time0 = light_time
    else:
        raise ValueError('light-travel time failed to converge')
    return tposition - cposition, tvelocity - cvelocity, t, light_time

def _jpl_name(target):
    if not isinstance(target, int):
        return type(target).__name__
    name = _jpl_code_name_dict.get(target)
    if name is None:
        return str(target)
    return '{0} {1}'.format(target, name)



# -*- coding: utf-8 -*-

import numpy as np

import jplephem
import sgp4
import skyfield
from skyfield.api import load
from skyfield.functions import load_bundled_npy

def main():
    print('Skyfield version: {0}'.format(skyfield.__version__))
    print('jplephem version: {0}'.format(jplephem.__version__))
    print('sgp4 version: {0}'.format(sgp4.__version__))

    ts = load.timescale()
    fmt = '%Y-%m-%d'

    final_leap = (ts._leap_tai[-1] - 1) / (24 * 60 * 60)
    print('Built-in leap seconds table ends with leap second at: {0}'
          .format(ts.tai_jd(final_leap).utc_strftime()))

    arrays = load_bundled_npy('iers.npz')
    daily_tt = arrays['tt_jd_minus_arange']
    daily_tt += np.arange(len(daily_tt))
    start = ts.tt_jd(daily_tt[0])
    end = ts.tt_jd(daily_tt[-1])
    print('Built-in ∆T table from finals2000A.all covers: {0} to {1}'
          .format(start.utc_strftime(fmt), end.utc_strftime(fmt)))

main()



import datetime as dt_module
import numbers
import numpy as np
import sys
from assay import assert_raises
from pytz import timezone
from numpy import array, inf
from skyfield import api
from skyfield.constants import DAY_S, T0
from skyfield.data import iers
from skyfield.functions import A
from skyfield.timelib import (
    GREGORIAN_START, GREGORIAN_START_ENGLAND, Time, Timescale,
    calendar_tuple, compute_calendar_date, julian_date, julian_day, utc,
)
from datetime import datetime

one_second = 1.0 / DAY_S
epsilon = one_second * 42.0e-6  # 20.1e-6 is theoretical best precision

continuous_timescale = ['tai', 'tt', 'tdb', 'ut1']
time_scale_name = ['utc', 'tai', 'tt', 'tdb', 'ut1']
time_value = [(1973, 1, 18, 1, 35, 37.5), 2441700.56640625]

def test_timescale_with_old_fashioned_leap_second_table():
    # Skyfield no longer uses the awkward old-style leap second table,
    # written before I knew better, with inf's on both ends; but in case
    # any users built such tables and supplied them manually:
    delta_t = array([[2442046.5005113888], [44.4782581]])
    leap_dates = array([-inf, 2441317.5, 2441499.5, 2441683.5, 2442048.5, inf])
    leap_offsets = array([10, 10, 10, 11, 12, 13])
    ts = Timescale(delta_t, leap_dates, leap_offsets)
    t = ts.tai(1973, 1, 1, 0, 0, [11, 12])
    assert t.utc.T.tolist() == [
        [1972, 12, 31, 23, 59, 60],
        [1973, 1, 1, 0, 0, 0],
    ]

    # The offset prior to the first leap second should be +10.
    t = ts.tai(1970, 1, 1)
    assert t.utc == (1969, 12, 31, 23, 59, 50)

def ts():
    yield api.load.timescale()

def ts_either():
    # This fixture is for "tests that should pass given *either* a
    # built-in Timescale or one loaded from a file."  Without a few such
    # tests, an adjustment to the loading scheme can introduce a bug
    # that no tests catch, because most tests use the builtin tables
    # intead of loading from a file.  The tests below that use this
    # fixture are ones that broke in the past when the loader code and
    # timescale disagreed about something.
    yield api.load.timescale()
    yield api.load.timescale(builtin=False)

def a(*args):
    return np.array(args)

def all_kinds_of_time_array(ts):
    yield ts.utc(2020, 10, [8, 9])
    yield ts.tai(2020, 10, [8, 9])
    yield ts.tt(2020, 10, [8, 9])
    yield ts.tdb(2020, 10, [8, 9])
    yield ts.ut1(2020, 10, [8, 9])

    jd = a(2459130.5, 2459131.5)

    yield ts.tai_jd(jd)
    yield ts.tt_jd(jd)
    yield ts.tdb_jd(jd)
    yield ts.ut1_jd(jd)
    yield Time(ts, jd)

    for jd, fraction in [
            (a(2459130.5, 2459131.5), 0.25),
            (2459130.5, a(0.0, 0.25)),
            (a(2459130.5, 2459131.5), a(0.0, 0.25)),
    ]:
        yield ts.tai_jd(jd, fraction)
        yield ts.tt_jd(jd, fraction)
        yield ts.tdb_jd(jd, fraction)
        # yield ts.ut1_jd(jd, fraction)  # not yet supported

    # We only support direct Time instantiation for the final case,
    # where jd and fraction are arrays that already agree in their
    # dimensions.

    yield Time(ts, jd, fraction)

def test_time_creation_methods(ts, continuous_timescale, time_value):
    method = getattr(ts, continuous_timescale)
    if isinstance(time_value, tuple):
        t = method(*time_value)
    else:
        t = method(jd=time_value) # TODO: deprecate
    assert getattr(t, continuous_timescale) == 2441700.56640625

    # Also go ahead and test the calendar and formatting operations.

    tup = getattr(t, continuous_timescale + '_calendar')()
    assert tup == (1973, 1, 18, 1, 35, 37.5)

    strftime = getattr(t, continuous_timescale + '_strftime')
    string = strftime()
    assert string == '1973-01-18 01:35:38 ' + continuous_timescale.upper()

    if sys.version_info <= (3,):
        return  # we do not currently support %f under Python 2

    string = strftime('%S.%f')
    assert string == '37.500000'

def test_now(ts):
    t = ts.now()
    assert isinstance(t, Time)

def test_months_overflow_correctly(ts):
    assert ts.tt(2020, -1).tt_strftime('%Y-%m') == '2019-11'
    assert ts.tt(2020, 15).tt_strftime('%Y-%m') == '2021-03'
    assert ts.tt(2020, [-1, 0, 1, 13, 14, 15]).tt_strftime('%Y-%m') == [
        '2019-11', '2019-12', '2020-01', '2021-01', '2021-02', '2021-03',
    ]

def test_days_overflow_correctly(ts):
    months = range(1, 13)
    assert ts.tt(2020, months, -1).tt_strftime('%Y-%m-%d') == [
        '2019-12-30', '2020-01-30', '2020-02-28', '2020-03-30',
        '2020-04-29', '2020-05-30', '2020-06-29', '2020-07-30',
        '2020-08-30', '2020-09-29', '2020-10-30', '2020-11-29',
    ]
    assert ts.tt(2020, months, 0).tt_strftime('%Y-%m-%d') == [
        '2019-12-31', '2020-01-31', '2020-02-29', '2020-03-31',
        '2020-04-30', '2020-05-31', '2020-06-30', '2020-07-31',
        '2020-08-31', '2020-09-30', '2020-10-31', '2020-11-30',
    ]
    assert ts.tt(2020, months, 32).tt_strftime('%Y-%m-%d') == [
        '2020-02-01', '2020-03-03', '2020-04-01', '2020-05-02',
        '2020-06-01', '2020-07-02', '2020-08-01', '2020-09-01',
        '2020-10-02', '2020-11-01', '2020-12-02', '2021-01-01',
    ]

def test_sequence_behaviors_of_time_that_is_not_array(ts):
    t = ts.utc(2023, 4, 2)
    with assert_raises(TypeError):
        for item in t:
            pass
    # The specific exception TypeError in the following case makes
    # Pandas inference.is_sequence() happy.
    with assert_raises(TypeError):
        len(t)

def test_time_can_be_indexed(ts):
    for t in all_kinds_of_time_array(ts):
        t[0]

def test_is_time_iterable(ts, time_scale_name):
    t = getattr(ts, time_scale_name)(2020, 9, (25, 26))
    for item in t:
        pass

def test_strftime_on_prehistoric_dates(ts_either):
    if sys.version_info <= (3,):
        return  # Python 2 time.strftime() complains about negative years

    ts = ts_either
    t = ts.tt(-746, 2, 26)
    assert t.utc_strftime('%Y %S') == '-746 18'
    assert t.ut1_strftime('%Y %S') == '-746 28'
    assert t.tai_strftime('%Y %S') == '-746 28'
    assert t.tt_strftime('%Y %S') == '-746 00'
    assert t.tdb_strftime('%Y %S') == '-746 00'

    t = ts.tt(-746, 2, [26, 26])
    assert t.utc_strftime('%Y %S') == ['-746 18'] * 2
    assert t.ut1_strftime('%Y %S') == ['-746 28'] * 2
    assert t.tai_strftime('%Y %S') == ['-746 28'] * 2
    assert t.tt_strftime('%Y %S') == ['-746 00'] * 2
    assert t.tdb_strftime('%Y %S') == ['-746 00'] * 2

def test_strftime_with_microseconds():
    if sys.version_info <= (3,):
        return  # we do not currently support %f under Python 2

    ts = api.load.timescale(builtin=False)  # load "ci/finals2000A.all"
    t = ts.tt(1980, 9, 12)
    assert t.utc_strftime('%Y %S %f') == '1980 08 816000'
    assert t.ut1_strftime('%Y %S %f') == '1980 08 892775'
    assert t.tai_strftime('%Y %S %f') == '1980 27 816000'
    assert t.tt_strftime('%Y %S %f') == '1980 00 000000'
    assert t.tdb_strftime('%Y %S %f') == '1980 59 998471'

    t = ts.tt(1980, 9, [12, 12])
    assert t.utc_strftime('%Y %S %f') == ['1980 08 816000'] * 2
    assert t.ut1_strftime('%Y %S %f') == ['1980 08 892775'] * 2
    assert t.tai_strftime('%Y %S %f') == ['1980 27 816000'] * 2
    assert t.tt_strftime('%Y %S %f') == ['1980 00 000000'] * 2
    assert t.tdb_strftime('%Y %S %f') == ['1980 59 998471'] * 2

def test_tai_fraction_loses_no_precision(ts):
    t = ts.tai_jd(2459008.0, 0.0123456789)
    assert t.whole == 2459008.0
    assert t.tai_fraction == 0.0123456789

def test_tdb_fraction_loses_no_precision(ts):
    t = ts.tdb_jd(2459008.0, 0.0123456789)
    assert t.whole == 2459008.0
    assert t.tdb_fraction == 0.0123456789

def test_tai_seconds_preserve_10_decimal_places_in_calendar_seconds(ts):
    t = ts.tai(2020, 6, 7, 2, 2, 12.0123456789)
    c = t.tai_calendar()
    assert c[:5] == (2020, 6, 7, 2, 2)
    assert '%.10f' % c[5] == '12.0123456789'

def test_tt_seconds_preserve_10_decimal_places_in_calendar_seconds(ts):
    t = ts.tt(2020, 6, 7, 2, 2, 12.0123456789)
    c = t.tt_calendar()
    assert c[:5] == (2020, 6, 7, 2, 2)
    assert '%.10f' % c[5] == '12.0123456789'

time_params_with_array = [
    ((2018, 2019, 2020), 3, 25, 13, 1, 10),
    (2018, (3, 4, 5), 25, 13, 1, 10),
    (2018, 3, (25, 26, 27), 13, 1, 10),
    (2018, 3, 25, (13, 14, 15), 1, 10),
    (2018, 3, 25, 13, (1, 2, 3), 10),
    (2018, 3, 25, 13, 1, (10, 11, 12)),
]

def test_time_creation_with_arrays(ts, time_scale_name, time_params_with_array):
    print(time_scale_name)
    t = getattr(ts, time_scale_name)(*time_params_with_array)
    t.utc_jpl()  # a reasonably complicated operation

def test_timescale_utc_method_with_array_inside(ts):
    seconds = np.arange(48.0, 58.0, 1.0)
    t = ts.utc(1973, 12, 29, 23, 59, seconds)
    assert seconds.shape == t.shape
    for i, second in enumerate(seconds):
        assert t.tai[i] == ts.utc(1973, 12, 29, 23, 59, second).tai

def test_that_building_time_from_naive_datetime_raises_exception(ts):
    with assert_raises(ValueError) as info:
        ts.from_datetime(datetime(1973, 12, 29, 23, 59, 48))
    assert 'import timezone' in str(info.exception)

def test_building_time_from_single_utc_datetime(ts):
    t = ts.from_datetime(datetime(1973, 12, 29, 23, 59, 48, tzinfo=utc))
    assert t.tai == 2442046.5
    t = ts.utc(datetime(1973, 12, 29, 23, 59, 48, tzinfo=utc))
    assert t.tai == 2442046.5

def test_building_time_from_single_utc_datetime_with_timezone(ts):
    tz = timezone('US/Eastern')
    t = ts.from_datetime(tz.localize(datetime(2020, 5, 10, 12, 44, 13, 797865)))
    dt, leap_second = t.utc_datetime_and_leap_second()
    assert dt == datetime(2020, 5, 10, 16, 44, 13, 797865, tzinfo=utc)
    assert leap_second == 0

def test_building_time_from_list_of_utc_datetimes(ts):
    datetimes = [
        datetime(1973, 12, 29, 23, 59, 48, tzinfo=utc),
        datetime(1973, 12, 30, 23, 59, 48, tzinfo=utc),
        datetime(1973, 12, 31, 23, 59, 48, tzinfo=utc),
        datetime(1974, 1, 1, 23, 59, 47, tzinfo=utc),
        datetime(1974, 1, 2, 23, 59, 47, tzinfo=utc),
        datetime(1974, 1, 3, 23, 59, 47, tzinfo=utc),
    ]
    t = ts.from_datetimes(datetimes)
    assert list(t.tai) == [
        2442046.5, 2442047.5, 2442048.5, 2442049.5, 2442050.5, 2442051.5,
    ]
    t = ts.utc(datetimes)
    assert list(t.tai) == [
        2442046.5, 2442047.5, 2442048.5, 2442049.5, 2442050.5, 2442051.5,
    ]

def test_building_time_from_python_date(ts):
    d = dt_module.date(2020, 7, 22)
    t = ts.utc(d)
    assert t.utc == (2020, 7, 22, 0, 0, 0.0)

def test_building_time_from_utc_julian_date(ts):
    t = ts._utc_jd(2457754.5, - one_second)
    assert t.utc == (2016, 12, 31, 23, 59, 59.0)  # no JD corresponds to s=60.0

    t = ts._utc_jd(2457754.5, 0.0)
    assert t.utc == (2017, 1, 1, 0, 0, 0.0)

    t = ts._utc_jd(2457754.5, one_second)
    assert t.utc == (2017, 1, 1, 0, 0, 1.0)

def test_utc_julian_date_accuracy(ts):
    if sys.version_info <= (3,):
        return  # our utc_strftime() hack does not work under Python 2

    t = ts._utc_jd(2460439.5, 0.36689744000250357)
    assert t.utc_strftime('%H:%M:%S.%f') == '08:48:19.938816'

def test_timescale_linspace(ts):
    t0 = ts.tt(2021, 11, 3, 6)
    t1 = ts.tt(2021, 11, 5, 18)
    t = ts.linspace(t0, t1, 3)
    assert [n for a in t.tt_calendar() for n in a] == [
	2021, 2021, 2021,
        11, 11, 11,
        3, 4, 5,
        6, 12, 18,
        0, 0, 0,
        0, 0, 0,
    ]

def test_converting_ut1_to_tt(ts):
    ten_thousand_years = 365 * 10000

    jd = api.T0 - ten_thousand_years
    t = ts.ut1(jd=jd)
    del t.ut1_fraction          # force re-computation of UT1
    print(jd - t.ut1)
    assert abs(jd - t.ut1) < 1e-10

    jd = api.T0 + ten_thousand_years
    t = ts.ut1(jd=jd)
    del t.ut1_fraction          # force re-computation of UT1
    print(jd - t.ut1)
    assert abs(jd - t.ut1) < 1e-10

def test_indexing_time(ts):
    t = ts.utc(1974, 10, range(1, 6))
    assert t.shape == (5,)
    t0 = t[0]
    assert t.tai[0] == t0.tai
    assert t.tt[0] == t0.tt
    assert t.tdb[0] == t0.tdb
    assert t.ut1[0] == t0.ut1
    assert t.delta_t[0] == t0.delta_t

def test_slicing_time(ts):
    t = ts.utc(1974, 10, range(1, 6))
    assert t.shape == (5,)
    t24 = t[2:4]
    assert t24.shape == (2,)
    assert (t.tai[2:4] == t24.tai).all()
    assert (t.tt[2:4] == t24.tt).all()
    assert (t.tdb[2:4] == t24.tdb).all()
    assert (t.ut1[2:4] == t24.ut1).all()
    assert (t.delta_t[2:4] == t24.delta_t).all()

def test_early_utc(ts_either):
    ts = ts_either
    t = ts.utc(1915, 12, 2, 3, 4, 5.6786786)
    assert abs(t.tt - 2420833.6283317441) < epsilon
    assert t.utc_iso() == '1915-12-02T03:04:06Z'

def test_astimezone(ts):
    t = ts.utc(1969, 7, 20, 20, 18)
    tz = timezone('US/Eastern')
    dt = t.astimezone(tz)
    assert dt == tz.localize(datetime(1969, 7, 20, 16, 18, 0, 0))

def test_astimezone_and_leap_second(ts):
    t = ts.utc(1969, 7, 20, 20, 18)
    tz = timezone('US/Eastern')
    dt, leap_second = t.astimezone_and_leap_second(tz)
    assert dt == tz.localize(datetime(1969, 7, 20, 16, 18, 0, 0))
    assert leap_second == 0

def test_toordinal(ts):
    t = ts.utc(1973, 12, 31, 11, 59, 60)
    assert t.toordinal() == 720623.5

def test_utc_datetime(ts):
    t = ts.utc(1969, 7, 20, 20, 18, 42.186479)
    dt = t.utc_datetime()
    assert dt == datetime(1969, 7, 20, 20, 18, 42, 186479, utc)

def test_utc_datetime_and_leap_second(ts):
    t = ts.utc(1969, 7, 20, 20, 18)
    dt, leap_second = t.utc_datetime_and_leap_second()
    assert dt == datetime(1969, 7, 20, 20, 18, 0, 0, utc)
    assert leap_second == 0

    t = ts.utc(1969, 7, 20, 20, [18, 19])
    dt, leap_second = t.utc_datetime_and_leap_second()
    assert list(dt) == [datetime(1969, 7, 20, 20, 18, 0, 0, utc),
                        datetime(1969, 7, 20, 20, 19, 0, 0, utc)]
    assert list(leap_second) == [0,0]

def test_utc_datetime_microseconds_round_trip(ts):
    dt = datetime(2020, 5, 10, 11, 50, 9, 727799, tzinfo=utc)
    t = ts.from_datetime(dt)
    dt2, leap_second = t.utc_datetime_and_leap_second()
    assert dt2 == dt
    assert leap_second == 0

def test_utc_datetime_agrees_with_public_utc_tuple(ts):
    # https://github.com/skyfielders/python-skyfield/issues/542
    # The %j day-of-year was advancing to the next day before strftime.
    t = ts.utc(2021, 1, 1, 23, 59, 59.9999798834251798497)
    assert t.utc[:5] == (2021, 1, 1, 23, 59)
    assert t.utc_strftime("%j") == '001'

def test_utc_datetime_exception_for_negative_year():
    ts = api.load.timescale()
    ts.julian_calendar_cutoff = GREGORIAN_START
    t = ts.utc(-1, 1, 1)
    with assert_raises(ValueError, 'negative years like the year -1'):
        t.utc_datetime()

    t = ts.utc([-2, -3, -1], 1, 1)
    with assert_raises(ValueError, 'negative years like the year -2'):
        t.utc_datetime()

def test_utc_datetime_exception_for_julian_leap_day():
    ts = api.load.timescale()
    ts.julian_calendar_cutoff = GREGORIAN_START
    t = ts.utc(700, 2, 29)
    with assert_raises(ValueError, 'Julian leap days like 700 February 29'):
        t.utc_datetime()

    t = ts.utc(700, 2, [28, 29, 30])
    with assert_raises(ValueError, 'Julian leap days like 700 February 29'):
        t.utc_datetime()

def test_iso_of_decimal_that_rounds_up(ts):
    t = ts.utc(1915, 12, 2, 3, 4, 5.6786786)
    assert t.utc_iso(places=0) == '1915-12-02T03:04:06Z'
    assert t.utc_iso(places=1) == '1915-12-02T03:04:05.7Z'
    assert t.utc_iso(places=2) == '1915-12-02T03:04:05.68Z'
    assert t.utc_iso(places=3) == '1915-12-02T03:04:05.679Z'
    assert t.utc_iso(places=4) == '1915-12-02T03:04:05.6787Z'

def test_iso_of_decimal_that_rounds_down(ts):
    t = ts.utc(2014, 12, 21, 6, 3, 1.234234)
    assert t.utc_iso(places=0) == '2014-12-21T06:03:01Z'
    assert t.utc_iso(places=1) == '2014-12-21T06:03:01.2Z'
    assert t.utc_iso(places=2) == '2014-12-21T06:03:01.23Z'
    assert t.utc_iso(places=3) == '2014-12-21T06:03:01.234Z'
    assert t.utc_iso(places=4) == '2014-12-21T06:03:01.2342Z'

def test_iso_of_leap_second_with_fraction(ts):
    t = ts.utc(1973, 12, 31, 23, 59, 60.12349)
    assert t.utc_iso(places=0) == '1973-12-31T23:59:60Z'
    assert t.utc_iso(places=1) == '1973-12-31T23:59:60.1Z'
    assert t.utc_iso(places=2) == '1973-12-31T23:59:60.12Z'
    assert t.utc_iso(places=3) == '1973-12-31T23:59:60.123Z'
    assert t.utc_iso(places=4) == '1973-12-31T23:59:60.1235Z'

def test_iso_of_array_showing_whole_seconds(ts):
    t = ts.utc(1973, 12, 31, 23, 59, np.arange(58.75, 63.1, 0.5))
    assert t.utc_iso(places=0) == [
        '1973-12-31T23:59:59Z',
        '1973-12-31T23:59:59Z',
        '1973-12-31T23:59:60Z',
        '1973-12-31T23:59:60Z',
        '1974-01-01T00:00:00Z',
        '1974-01-01T00:00:00Z',
        '1974-01-01T00:00:01Z',
        '1974-01-01T00:00:01Z',
        '1974-01-01T00:00:02Z',
        ]

def test_iso_of_array_showing_fractions(ts):
    t = ts.utc(1973, 12, 31, 23, 59, np.arange(58.75, 63.1, 0.5))
    assert t.utc_iso(places=2) == [
        '1973-12-31T23:59:58.75Z',
        '1973-12-31T23:59:59.25Z',
        '1973-12-31T23:59:59.75Z',
        '1973-12-31T23:59:60.25Z',
        '1973-12-31T23:59:60.75Z',
        '1974-01-01T00:00:00.25Z',
        '1974-01-01T00:00:00.75Z',
        '1974-01-01T00:00:01.25Z',
        '1974-01-01T00:00:01.75Z',
    ]

def test_jpl_format(ts):
    t = ts.utc(range(-300, 301, 100), 7, 1)
    assert t.utc_jpl() == [
        'B.C. 0301-Jul-01 00:00:00.0000 UTC',
        'B.C. 0201-Jul-01 00:00:00.0000 UTC',
        'B.C. 0101-Jul-01 00:00:00.0000 UTC',
        'B.C. 0001-Jul-01 00:00:00.0000 UTC',
        'A.D. 0100-Jul-01 00:00:00.0000 UTC',
        'A.D. 0200-Jul-01 00:00:00.0000 UTC',
        'A.D. 0300-Jul-01 00:00:00.0000 UTC',
        ]

def test_strftime_of_a_leap_second(ts):
    t = ts.utc(1973, 12, 31, 23, 59, 60)
    assert t.utc_strftime('%Y %m %d %H %M %S') == '1973 12 31 23 59 60'

def test_strftime_of_date_array_over_a_leap_second(ts):
    t = ts.utc(1973, 12, 31, 23, 59, np.arange(59.0, 61.1, 1.0))
    assert t.utc_strftime('%a %Y %m %d %H %M %S') == [
        'Mon 1973 12 31 23 59 59',
        'Mon 1973 12 31 23 59 60',
        'Tue 1974 01 01 00 00 00',
    ]

def test_strftime_day_of_year(ts):
    # Based on example date at https://strftime.org/
    assert ts.utc(2013, 9, 29).utc_strftime('%j') == '272'
    assert ts.utc(2013, 9, 30).utc_strftime('%j') == '273'
    assert ts.utc(2013, 9, 30, 23, 59).utc_strftime('%j') == '273'
    assert ts.utc(2013, 9, 30, 23, 60).utc_strftime('%j') == '274'

    assert ts.utc(2013, 9, [29, 30]).utc_strftime('%j') == ['272', '273']

def test_leap_second(ts):

    # During 1973 the offset between UTC and TAI was 12.0 seconds, so
    # TAI should reach the first moment of 1974 while the UTC clock is
    # still reading 12s before midnight (60 - 12 = 48).  Happily, the
    # fraction 0.5 can be precisely represented in floating point, so we
    # can use a bare `==` in this assert:

    t0 = ts.utc(1973, 12, 31, 23, 59, 48.0).tai
    assert t0 == 2442048.5

    # Here are some more interesting values:

    t1 = ts.utc(1973, 12, 31, 23, 59, 58.0).tai
    t2 = ts.utc(1973, 12, 31, 23, 59, 59.0).tai
    t3 = ts.utc(1973, 12, 31, 23, 59, 60.0).tai
    t4 = ts.utc(1974, 1, 1, 0, 0, 0.0).tai
    t5 = ts.utc(1974, 1, 1, 0, 0, 1.0).tai

    # The step from 23:59:59 to 0:00:00 is here a two-second step,
    # because of the leap second 23:59:60 that falls in between:

    assert abs(t4 - t2 - 2.0 * one_second) < epsilon

    # Otherwise, the five dates given above are all one second apart:

    assert abs(t2 - t1 - one_second) < epsilon
    assert abs(t3 - t2 - one_second) < epsilon
    assert abs(t4 - t3 - one_second) < epsilon
    assert abs(t5 - t4 - one_second) < epsilon

    # And all these dates can be converted back to UTC.

    assert ts.tai(jd=t0).utc_iso() == '1973-12-31T23:59:48Z'
    assert ts.tai(jd=t1).utc_iso() == '1973-12-31T23:59:58Z'
    assert ts.tai(jd=t2).utc_iso() == '1973-12-31T23:59:59Z'
    assert ts.tai(jd=t3).utc_iso() == '1973-12-31T23:59:60Z'
    assert ts.tai(jd=t4).utc_iso() == '1974-01-01T00:00:00Z'
    assert ts.tai(jd=t5).utc_iso() == '1974-01-01T00:00:01Z'

def test_leap_second_sensitivity(ts):
    t = ts.utc(2017, 1, 1, 0, 0, 0)
    a, b = t.whole, t.tai_fraction

    # First, make sure the UTC time round-trips.
    t2 = ts.tai_jd(a, b)
    tup = t2._utc_tuple(0.0)
    assert tup == (2017, 1, 1, 0, 0, 0.0)

    # Second, bump back infinitesimally into the previous leap second
    # and make sure the UTC time is just shy of second 61.0.
    b = np.nextafter(b, -1)
    t3 = ts.tai_jd(a, b)
    tup = t3._utc_tuple(0.0)
    assert tup[:5] == (2016, 12, 31, 23, 59)
    assert tup[5] > 60.99999999999

def test_delta_t(ts):
    # The IERS "finals2000A.all" for 2000 Jan 1 gives DUT1 = 0.3554779,
    # and 0.3554779 - 0.184 - 1.0 = -0.8285221.
    t = ts.utc(2000, 1, 1, 0, 0, 0)
    assert t.delta_t == 63.8285221

    # Check historic value. Compare to the table in Morrison and
    # Stephenson 2004, the tolerance is 2 sigma
    t = ts.utc(year=1000)
    assert abs(t.delta_t - 1570.0) < 110.0

    # Check far-future value against the long-term parabola formula.
    centuries = 20
    t = ts.J(1825 + centuries * 100)
    assert t.delta_t == -320 + 32.5 * centuries**2

def test_dut1(ts):
    # Roughly agreeing with tables on NIST website
    t = ts.utc(2017, 3, 30)
    assert str(t.dut1)[:3] == '0.4'

    t = ts.utc(2018, 9, 21)
    assert str(t.dut1)[:3] == '0.0'

    t = ts.utc(2019, 5, 2)
    assert str(t.dut1)[:4] == '-0.1'  # table says -0.2

def test_polar_motion_table():
    with api.load.open('finals2000A.all') as f:
        finals_data = iers.parse_x_y_dut1_from_finals_all(f)

    ts = api.load.timescale()
    iers.install_polar_motion_table(ts, finals_data)

    bump = 1e-4  # TODO: can this be improved?

    t = ts.utc(1973, 1, 3, 0, 0, -bump)
    sprime, x, y = t.polar_motion_angles()
    assert x > 0.118980
    assert y > 0.135656

    t = ts.utc(1973, 1, 3, 0, 0, +bump)
    sprime, x, y = t.polar_motion_angles()
    assert x < 0.118980
    assert y < 0.135656

def test_J(ts):
    assert ts.J(2000).tt == T0
    assert ts.J(1900).tt == T0 - 36525.0
    assert (ts.J([1900, 2000]).tt == [T0 - 36525.0, T0]).all()
    assert ts.tt(2000, 1, 1.5).J == 2000.0
    assert ts.tt(1900, 1, 0.5).J == 1900.0

def test_time_repr(ts):

    # Check that repr return is a str (this is required on Python 2,
    # unicode is not allowed)
    assert isinstance(repr(ts.utc(year=2000)), str)

    # Check array conversion
    assert isinstance(repr(ts.utc(year=range(2000, 2010))), str)

    assert repr(ts.tt_jd(1)) == '<Time tt=1.0>'
    assert repr(ts.tt_jd([])) == '<Time tt=[]>'
    assert repr(ts.tt_jd([1])) == '<Time tt=[1.]>'
    assert repr(ts.tt_jd([1, 2])) == '<Time tt=[1. 2.]>'
    assert repr(ts.tt_jd([1, 2, 3])) == '<Time tt=[1. 2. 3.]>'
    assert repr(ts.tt_jd([1, 2, 3, 4])) == '<Time tt=[1.0 ... 4.0] len=4>'

def test_jd_calendar():
    # Check a specific instance (using UNIX epoch here, though that's
    # an arbitrary choice)
    jd_unix = 2440587.5
    cal_unix = (1970, 1, 1, 0, 0, 0.0)

    cal = calendar_tuple(jd_unix)

    # Check that the returned value is correct
    assert cal == cal_unix

    # Check that all the return types are correct
    assert isinstance(cal[0], numbers.Integral)  # Year
    assert isinstance(cal[1], numbers.Integral)  # Month
    assert isinstance(cal[2], numbers.Integral)  # Day
    assert isinstance(cal[3], numbers.Integral)  # Hour
    assert isinstance(cal[4], numbers.Integral)  # Minute
    assert isinstance(cal[5], numbers.Real)  # Second

    # Check backward conversion
    assert julian_date(*cal) == jd_unix

    # Check array conversion components
    jd_array = jd_unix + np.arange(5.0)
    cal_array = calendar_tuple(jd_array)

    assert (cal_array[0] == 1970).all()
    assert (cal_array[1] == 1).all()
    assert (cal_array[2] == np.arange(1, 6)).all()
    assert (cal_array[3] == 0).all()
    assert (cal_array[4] == 0).all()
    assert (cal_array[5] == 0.0).all()

    # Check reversal of array
    assert (julian_date(*cal_array) == jd_array).all()

def test_raw_julian_gregorian_cutover():
    gregory = 2299161
    assert compute_calendar_date(gregory - 2, gregory) == (1582, 10, 3)
    assert compute_calendar_date(gregory - 1, gregory) == (1582, 10, 4)
    assert compute_calendar_date(gregory + 0, gregory) == (1582, 10, 15)
    assert compute_calendar_date(gregory + 1, gregory) == (1582, 10, 16)

    jd = np.arange(gregory - 2, gregory + 2)
    assert [list(a) for a in compute_calendar_date(jd, gregory)] == [
        [1582, 1582, 1582, 1582],
        [10, 10, 10, 10],
        [3, 4, 15, 16],
    ]

    assert julian_day(1582, 10, 3, gregory) == (gregory - 2)
    assert julian_day(1582, 10, 4, gregory) == (gregory - 1)
    assert julian_day(1582, 10, 15, gregory) == (gregory + 0)
    assert julian_day(1582, 10, 16, gregory) == (gregory + 1)

    days = [3, 4, 15, 16]
    assert list(julian_day(1582, 10, np.array(days), gregory)) == [
        2299159, 2299160, 2299161, 2299162,
    ]

def test_constructor_julian_gregorian_cutover(time_scale_name):
    if sys.version_info <= (3,):
        return  # Python 2 time.strftime() complains about the year 1582

    def jd(y, m, d):
        t = getattr(ts, time_scale_name)(y, m, d)
        if time_scale_name == 'utc':
            return sum(t._utc_seconds(0.0)) / DAY_S
        return getattr(t, time_scale_name)

    ts = api.load.timescale()

    assert jd(1582, 10, 4) == 2299149.5
    assert jd(1582, 10, 15) == 2299160.5
    assert jd(1752, 9, 2) == 2361209.5
    assert jd(1752, 9, 14) == 2361221.5

    ts.julian_calendar_cutoff = GREGORIAN_START

    assert jd(1582, 10, 4) == 2299159.5
    assert jd(1582, 10, 15) == 2299160.5
    assert jd(1752, 9, 2) == 2361209.5
    assert jd(1752, 9, 14) == 2361221.5

    ts.julian_calendar_cutoff = GREGORIAN_START_ENGLAND

    assert jd(1582, 10, 4) == 2299159.5
    assert jd(1582, 10, 15) == 2299170.5
    assert jd(1752, 9, 2) == 2361220.5
    assert jd(1752, 9, 14) == 2361221.5

def test_calendar_tuple_julian_gregorian_cutover(time_scale_name):
    if sys.version_info <= (3,):
        return  # Python 2 time.strftime() complains about the year 1582

    def ymd(jd):
        t = ts.tt_jd(jd, 0.1)
        if time_scale_name == 'utc':
            return t.utc[:3]
        return getattr(t, time_scale_name + '_calendar')()[:3]

    ts = api.load.timescale()

    assert ymd(2299149.5) == (1582, 10, 4)
    assert ymd(2299160.5) == (1582, 10, 15)
    assert ymd(2361209.5) == (1752, 9, 2)
    assert ymd(2361221.5) == (1752, 9, 14)

    ts.julian_calendar_cutoff = GREGORIAN_START

    assert ymd(2299159.5) == (1582, 10, 4)
    assert ymd(2299160.5) == (1582, 10, 15)
    assert ymd(2361209.5) == (1752, 9, 2)
    assert ymd(2361221.5) == (1752, 9, 14)

    ts.julian_calendar_cutoff = GREGORIAN_START_ENGLAND

    assert ymd(2299159.5) == (1582, 10, 4)
    assert ymd(2299170.5) == (1582, 10, 15)
    assert ymd(2361220.5) == (1752, 9, 2)
    assert ymd(2361221.5) == (1752, 9, 14)

def test_strftime_julian_gregorian_cutover(time_scale_name):
    if sys.version_info <= (3,):
        return  # Python 2 time.strftime() complains about the year 1582

    def ymd(jd):
        t = ts.tt_jd(jd, 0.1)
        return getattr(t, time_scale_name + '_strftime')('%Y %m %d')

    ts = api.load.timescale()

    assert ymd(2299149.5) == '1582 10 04'
    assert ymd(2299160.5) == '1582 10 15'
    assert ymd(2361209.5) == '1752 09 02'
    assert ymd(2361221.5) == '1752 09 14'

    ts.julian_calendar_cutoff = GREGORIAN_START

    assert ymd(2299159.5) == '1582 10 04'
    assert ymd(2299160.5) == '1582 10 15'
    assert ymd(2361209.5) == '1752 09 02'
    assert ymd(2361221.5) == '1752 09 14'

    ts.julian_calendar_cutoff = GREGORIAN_START_ENGLAND

    assert ymd(2299159.5) == '1582 10 04'
    assert ymd(2299170.5) == '1582 10 15'
    assert ymd(2361220.5) == '1752 09 02'
    assert ymd(2361221.5) == '1752 09 14'

def test_time_equality(ts):
    t0 = ts.tt_jd(2459008.5, 0.125)
    t1 = ts.tt_jd(2459008.0, 0.625)
    assert t0 == t1
    assert t1 - t0 == 0.0
    assert hash(t0) == hash(t1)

    t2 = ts.tt_jd(2459008.0, 0.6251)
    assert t2 != t0
    assert t2 - t0 > 0
    assert hash(t0) != hash(t2)

def test_time_math(ts):
    t = ts.tt_jd(2459008.5, 0.125)

    assert (t - 1).tt_strftime() == '2020-06-07 03:00:00 TT'
    assert (t + 1).tt_strftime() == '2020-06-09 03:00:00 TT'

    assert (t - 1.25).tt_strftime() == '2020-06-06 21:00:00 TT'
    assert (t + 1.25).tt_strftime() == '2020-06-09 09:00:00 TT'

    bump = dt_module.timedelta(days=1, seconds=1)
    assert (t - bump).tt_strftime() == '2020-06-07 02:59:59 TT'
    assert (t + bump).tt_strftime() == '2020-06-09 03:00:01 TT'

    bump = dt_module.timedelta(microseconds=300)
    assert (t - bump).utc_jpl() == 'A.D. 2020-Jun-08 02:58:50.8157 UTC'
    assert (t + bump).utc_jpl() == 'A.D. 2020-Jun-08 02:58:50.8163 UTC'

    bump = A[1.0, 2.0]  # TODO: make this work with one element
    assert tuple((t - bump).tt) == (2459007.625, 2459006.625)
    assert tuple((t + bump).tt) == (2459009.625, 2459010.625)

    assert t < t + 1



"""Formulae for specific earth behaviors and effects."""

from numpy import (abs, arcsin, arccos, arctan2, array, clip, cos, minimum,
                   nan_to_num, pi, sin, sqrt, tan, where, zeros_like)

from .constants import (AU_M, ANGVEL, DAY_S, DEG2RAD, ERAD,
                        IERS_2010_INVERSE_EARTH_FLATTENING, RAD2DEG, T0, tau)
from .functions import dots

earth_radius_au = ERAD / AU_M
one_minus_flattening = 1.0 - 1.0 / IERS_2010_INVERSE_EARTH_FLATTENING
one_minus_flattening_squared = one_minus_flattening * one_minus_flattening

def terra(latitude, longitude, elevation, gast):
    """Deprecated conversion from lat,lon,t -> GCRS; neglects polar motion."""
    zero = zeros_like(gast)
    sinphi = sin(latitude)
    cosphi = cos(latitude)
    c = 1.0 / sqrt(cosphi * cosphi +
                   sinphi * sinphi * one_minus_flattening_squared)
    s = one_minus_flattening_squared * c
    ach = earth_radius_au * c + elevation
    ash = earth_radius_au * s + elevation

    # Compute local sidereal time factors at the observer's longitude.

    stlocl = 15.0 * DEG2RAD * gast + longitude
    sinst = sin(stlocl)
    cosst = cos(stlocl)

    # Compute position vector components in au.

    ac = ach * cosphi
    acsst = ac * sinst
    accst = ac * cosst
    pos = array((accst, acsst, zero + ash * sinphi))

    # Compute velocity vector components in au/day.

    vel = ANGVEL * DAY_S * array((-acsst, accst, zero))

    return pos, vel

def reverse_terra(xyz_au, gast, iterations=3):
    """Deprecated conversion from GCRS -> lat,lon,t; neglects polar motion."""
    x, y, z = xyz_au
    R = sqrt(x*x + y*y)

    lon = (arctan2(y, x) - 15 * DEG2RAD * gast - pi) % tau - pi
    lat = arctan2(z, R)

    a = ERAD / AU_M
    f = 1.0 / IERS_2010_INVERSE_EARTH_FLATTENING
    e2 = 2.0*f - f*f
    i = 0
    C = 1.0
    while i < iterations:
        i += 1
        C = 1.0 / sqrt(1.0 - e2 * (sin(lat) ** 2.0))
        lat = arctan2(z + a * C * e2 * sin(lat), R)
    elevation_m = ((R / cos(lat)) - a * C) * AU_M
    return lat, lon, elevation_m

def compute_limb_angle(position_au, observer_au):
    """Determine the angle of an object above or below the Earth's limb.

    Given an object's GCRS `position_au` |xyz| vector and the position
    of an `observer_au` as a vector in the same coordinate system,
    return a tuple that provides `(limb_ang, nadir_ang)`:

    limb_angle
        Angle of observed object above (+) or below (-) limb in degrees.
    nadir_angle
        Nadir angle of observed object as a fraction of apparent radius
        of limb: <1.0 means below the limb, =1.0 means on the limb, and
        >1.0 means above the limb.

    """
    # Compute the distance to the object and the distance to the observer.

    disobj = sqrt(dots(position_au, position_au))
    disobs = sqrt(dots(observer_au, observer_au))

    # Compute apparent angular radius of Earth's limb.

    aprad = arcsin(minimum(earth_radius_au / disobs, 1.0))

    # Compute zenith distance of Earth's limb.

    zdlim = pi - aprad

    # Compute zenith distance of observed object.

    coszd = dots(position_au, observer_au) / (disobj * disobs)
    coszd = clip(coszd, -1.0, 1.0)
    zdobj = arccos(coszd)

    # Angle of object wrt limb is difference in zenith distances.

    limb_angle = (zdlim - zdobj) * RAD2DEG

    # Nadir angle of object as a fraction of angular radius of limb.

    nadir_angle = (pi - zdobj) / aprad

    return limb_angle, nadir_angle

def sidereal_time(t):
    """Compute Greenwich Mean Sidereal Time (GMST) in hours at time ``t``."""

    theta = earth_rotation_angle(t.whole, t.ut1_fraction)

    # The equinox method.  See Circular 179, Section 2.6.2.
    # Precession-in-RA terms in mean sidereal time taken from third
    # reference, eq. (42), with coefficients in arcseconds.

    t = (t.whole - T0 + t.tdb_fraction) / 36525.0
    st =        ( 0.014506 +
        (((( -    0.0000000368   * t
             -    0.000029956  ) * t
             -    0.00000044   ) * t
             +    1.3915817    ) * t
             + 4612.156534     ) * t)

    return (st / 54000.0 + theta * 24.0) % 24.0

def earth_rotation_angle(jd_ut1, fraction_ut1=0.0):
    """Return the value of the Earth Rotation Angle (theta) for a UT1 date.

    Uses the expression from the note to IAU Resolution B1.8 of 2000.
    Returns a fraction between 0.0 and 1.0 whole rotations.

    """
    th = 0.7790572732640 + 0.00273781191135448 * (jd_ut1 - T0 + fraction_ut1)
    return (th % 1.0 + jd_ut1 % 1.0 + fraction_ut1) % 1.0

def refraction(alt_degrees, temperature_C, pressure_mbar):
    """Given an observed altitude, estimate atmospheric refraction, in degrees.

    Zero refraction is returned both for objects very near the zenith,
    as well as for objects more than one degree below the horizon.

    """
    r = 0.016667 / tan((alt_degrees + 7.31 / (alt_degrees + 4.4)) * DEG2RAD)
    d = r * (0.28 * pressure_mbar / (temperature_C + 273.0))
    return where((-1.0 <= alt_degrees) & (alt_degrees <= 89.9), d, 0.0)

def refract(alt_degrees, temperature_C, pressure_mbar):
    """Given an unrefracted `alt` determine where it will appear in the sky."""
    alt = alt_degrees
    while True:
        alt1 = alt
        alt = alt_degrees + refraction(alt, temperature_C, pressure_mbar)
        converged = nan_to_num(abs(alt - alt1)).max() <= 3.0e-5
        if converged:
            break
    return alt



from skyfield.curvelib import Splines, build_spline_given_ends

def test_spline_and_derivative():
    parameters = 10, 12, 2.0, 3.0, 5.0, 7.0
    x = 8.0, 9.0, 10.0, 11.0, 12.0
    expected_values = 3.0, 5.0, 7.0, 10.5, 17.0
    expected_slope = 2.5, 1.75, 2.5, 4.75, 8.5

    # Does the Spline class work?

    curve = Splines(parameters)
    assert tuple(curve(x)) == expected_values
    assert tuple(curve.derivative(x)) == expected_slope

    # Can we rebuild the parameters from its endpoints and slopes?

    parameters2 = build_spline_given_ends(
        10, curve(10), curve.derivative(10),
        12, curve(12), curve.derivative(12))

    assert parameters == parameters2

    # If we rebuild the spline around different bounds, does it still
    # produce the same curve?

    for lower, upper in (8,10), (8,9), (21,23):
        parameters3 = build_spline_given_ends(
            lower, curve(lower), curve.derivative(lower),
            upper, curve(upper), curve.derivative(upper))
        curve3 = Splines(parameters3)
        assert tuple(curve3(x)) == expected_values
        assert tuple(curve3.derivative(x)) == expected_slope



from skyfield.api import load
from skyfield import eclipselib

def test_lunar_eclipses():
    # The documentation test already confirms the dates of these two
    # eclipses; here, we confirm that the data structures all match.
    ts = load.timescale()
    eph = load('de421.bsp')

    t0 = ts.utc(2019, 1, 1)
    t1 = ts.utc(2020, 1, 1)
    t, y, details = eclipselib.lunar_eclipses(t0, t1, eph)

    assert len(t) == len(y) == 2
    for name, item in details.items():
        assert len(item) == len(t)
